"""
Module with functions for generate chunks
"""

# pylint: disable=inconsistent-return-statements
from collections.abc import Generator
from typing import Any

from .split import _validate_type


def chunks(iterable: list[Any] | tuple[Any, ...], size_chunk: int) -> Generator:
    """
    Method for splitting into many parts with len size_chunk generator
    :param iterable: object to split
    :type iterable: Union[List[Any], Tuple[Any, ...]]
    :param size_chunk: size_chunk
    :type size_chunk: int
    :return: generator of split iterable object len generator equal cnt variable
    :rtype: Generator
    :raises TypeError: if incorrect type of `iterable` or `size_chunk` variables
    :raises ValueError: if `size_chunk` variable is not in a valid value
    """
    if not isinstance(size_chunk, int):
        raise TypeError("Incorrect type `size_chunk`")
    if size_chunk < 1:
        raise ValueError("Chunk size must be >= 1")
    _validate_type(iterable=iterable)
    for i in range(0, len(iterable), size_chunk):
        yield iterable[i : i + size_chunk]



"""
Module with function for convert bytes size to correct size in kb/mb/gb
"""

# pylint: disable=super-with-arguments
from __future__ import annotations

from enum import StrEnum
import math
from numbers import Number
from typing import Any

from pydantic import BaseModel, Field


class SystemTypes(StrEnum):
    """
    Enum of available conversion systems
    """

    SI = "si_name"
    IEC = "iec_name"
    VERBOSE = "verbose_name"
    ALTERNATIVE = "alternative_name"
    TRADITIONAL = "traditional_name"


class SystemValue(BaseModel):
    """
    Class for storing information about different weights
    """

    pow: int = Field(
        ...,
        description="the power to which you need to raise the bytes "
        "for the correct calculation",
    )
    traditional_name: str = Field(..., description="Traditional name for this type")
    alternative_name: str | tuple[str, str] = Field(
        ..., description="Alternative name for this type"
    )
    verbose_name: tuple[str, str] = Field(
        ..., description="Full name without abbreviations"
    )
    iec_name: str = Field(..., description="IEC system name")
    si_name: str = Field(None, description="Si system name")

    value_base: int = Field(
        None, description="The number of bytes in one unit of this type"
    )
    value_si: int = Field(
        None, description="The number of bytes in one unit of this type in system si"
    )

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.value_base = 1024**self.pow
        self.value_si = 1000**self.pow
        if not self.si_name:
            self.si_name = self.traditional_name

    def get_size_suffix(
        self, system: str = "traditional"
    ) -> str | tuple[str, str] | None:
        """
        method return str title for chose type
        :param system: title system for view
         allowed: [si, iec, verbose, alternative, traditional]
        :type system: str
        :return: names for chose system
        :rtype: str | tuple[str, str]
        :raise ValueError: if there is no selected system
        """
        system = system.upper()
        try:
            name_key = SystemTypes[system].value
            return self.__getattribute__(name_key)
        except KeyError:
            raise ValueError(
                f"Incorrect name system."
                f" Allowed names: {[i.name for i in SystemTypes]}"
            ) from BaseException

    def get_amount(
        self,
        bytes_value: int,
        round_to: int = 2,
        is_si: bool = False,
    ) -> float:
        """
        Method get size of bytes value in system
        :param bytes_value:  bytes to convert
        :type bytes_value: int
        :param round_to: decimal point rounding precision
        :type round_to:  int
        :param is_si: is to 'si' system convert
        :type is_si: bool
        :return:
        """
        if is_si:
            val = self.value_si
        else:
            val = self.value_base
        return round(bytes_value / val, round_to)

    def get_normalized_size(
        self,
        bytes_value: int,
        system: str | None = "traditional",
        round_to: int | None = 2,
    ) -> str:
        """
        Method convert bytes value in chose system
        :param bytes_value: bytes value to convert
        :type bytes_value: int
        :param system: system to which to convert
        :type system: str
        :param round_to:
        :type round_to: decimal point rounding precision
        :return: converted value
        :rtype: str
        :raise ValueError: if incorrect system name
        """
        system = system.upper()
        is_si = system == "SI"
        amount = self.get_amount(
            bytes_value=bytes_value, round_to=round_to, is_si=is_si
        )
        suffix_names = self.get_size_suffix(system=system)

        if isinstance(suffix_names, tuple):
            singular, multiple = suffix_names
            if amount == 1.0:
                suffix_amount = singular
            else:
                suffix_amount = multiple
        else:
            suffix_amount = suffix_names
        frac, whole = math.modf(amount)
        if frac == 0.0:
            amount = int(whole)
        return f"{amount}{suffix_amount}"

    def __lt__(self, other: Number | SystemValue) -> bool:
        if isinstance(other, SystemValue):
            result = self.value_base < other.value_base
        elif isinstance(other, Number):
            result = self.value_base < other  # type: ignore
        else:
            raise TypeError("incorrect value to compare")
        return result

    def __le__(self, other: Number | SystemValue) -> bool:
        return self < other or self == other

    def __gt__(self, other: Number | SystemValue) -> bool:
        result: bool = False
        if isinstance(other, SystemValue):
            result = self.value_base > other.value_base
        elif isinstance(other, Number):
            result = self.value_base > other  # type: ignore
        else:
            raise TypeError("incorrect value type to compare")
        return result

    def __eq__(self, other: Number | SystemValue) -> bool:  # type: ignore
        result: bool = False
        if isinstance(other, SystemValue):
            result = super().__eq__(other)
        elif isinstance(other, Number):
            result = self.value_base == other
        else:
            raise TypeError("incorrect value type to compare")
        return result

    def __ge__(self, other: Number | SystemValue) -> bool:
        return self > other or self == other


LIST_SYSTEMS = [
    SystemValue(
        pow=5,
        traditional_name="P",
        alternative_name="PB",
        verbose_name=(" petabyte", " petabytes"),
        iec_name="Pi",
    ),
    SystemValue(
        pow=4,
        traditional_name="T",
        alternative_name="TB",
        verbose_name=(" terabyte", " terabytes"),
        iec_name="Ti",
    ),
    SystemValue(
        pow=3,
        traditional_name="G",
        alternative_name="GB",
        verbose_name=(" gigabyte", " gigabytes"),
        iec_name="Gi",
    ),
    SystemValue(
        pow=2,
        traditional_name="M",
        alternative_name="MB",
        verbose_name=(" megabyte", " megabytes"),
        iec_name="Mi",
    ),
    SystemValue(
        pow=1,
        traditional_name="K",
        alternative_name="KB",
        verbose_name=(" kilobyte", " kilobytes"),
        iec_name="Ki",
    ),
    SystemValue(
        pow=0,
        traditional_name="B",
        alternative_name=(" byte", " bytes"),
        verbose_name=(" byte", " bytes"),
        iec_name="",
    ),
]


def size(
    bytes_value: int, system: str | None = "traditional", round_to: int = 2
) -> str:
    """
    Method for print size of file from bytes to pretty print
    :param bytes_value: bytes to convert
    :type bytes_value:
    :param system:
    :type system:
    :param round_to: decimal point rounding precision
    :type round_to: int
    :return: converted bytes_value to pretty string
    :rtype: str
    :raise ValueError: if incorrect system name or round_to is less 0
    :raise TypeError: if incorrect type round_to

    :example:
    Using the traditional system, where a factor of 1024 is used::

    >>> size(10)
    '10B'
    >>> size(2000)
    '1.95K'
    >>> size(20000)
    '19.53K'
    >>> size(200000)
    '0.19M'
    >>> size(1000000)
    '0.95M'
    >>> size(2000000)
    '1.91M'
    # use si
    >>> size(10, system="si")
    '10B'
    >>> size(2000, system="si")
    '2K'
    >>> size(20000, system="si")
    '20K'
    >>> size(200000, system="si")
    '0.2M'
    >>> size(1000000, system="si")
    '1M'
    >>> size(2000000, system="si")
    '2M'
    """
    if not isinstance(round_to, int):
        raise TypeError()
    if round_to < 1:
        raise ValueError()
    if not system:
        system = "traditional"
    system = system.upper()
    if system not in [i.name for i in SystemTypes]:
        raise ValueError(
            f"Incorrect system name - `{system}`."
            f" Allowed systems {[i.name for i in SystemTypes]}"
        )
    factor = None  # type: SystemValue|None
    for _factor in LIST_SYSTEMS:  # pragma: no cover
        if bytes_value >= _factor.value_base / 10:
            factor = _factor
            break

    return factor.get_normalized_size(
        bytes_value=bytes_value, system=system, round_to=round_to
    )



"""
Module with functions for split iterable objects
"""

# pylint: disable=inconsistent-return-statements
from collections.abc import Generator, Iterable
from typing import Any


def _validate_type(iterable: list[Any] | tuple[Any, ...]) -> type | None:
    """
    Validate the type of variable iterable
    :param iterable: Iterable type to variable
    :type iterable: Union[List[Any], Tuple[Any, ...]]
    :return: if correct type variable `iterable` than return type
    :rtype: type | None
    :raise TypeError: if incorrect type `iterable`
    """
    if not isinstance(iterable, (list, tuple)):
        raise TypeError(
            f"Incorrect type variable `iterable`."
            f" Valid types are [list, tuple] and not {type(iterable)}"
        )
    return type(iterable)


def split(iterable: list[Any] | tuple[Any, ...], cnt: int) -> Generator:
    """
    Method for splitting into equal cnt parts generator
    :param iterable: object to split
    :type iterable: Union[List[Any], Tuple[Any, ...]]
    :param cnt: cnt partitions
    :type cnt: int
    :return: generator of split iterable object len generator equal cnt variable
    :rtype: Generator
    :raises TypeError: if incorrect type of `iterable` variable
    """
    _validate_type(iterable=iterable)
    k, j = divmod(len(iterable), cnt)  # type: int, int
    for i in range(cnt):
        yield iterable[i * k + min(i, j) : (i + 1) * k + min(i + 1, j)]


def split_as_iterable(
    iterable: list[Any] | tuple[Any, ...], cnt: int
) -> list[list[Any] | tuple[tuple[Any]]]:
    """
    Method for splitting into equal parts
    :param iterable: object to split
    :type iterable: Union[List[Any], Tuple[Any, ...]]
    :param cnt: cnt partitions
    :type cnt: int
    :return: iterable object with partitions
    :rtype: Union[List[Any], Tuple[Any, ...]]
    :raises TypeError: if incorrect type of `iterable` variable
    """
    _current_type = _validate_type(iterable)  # type: type
    return _current_type(split(iterable=iterable, cnt=cnt))


def split_with_overlap(
    data: list, n_parts: int = 10, overlap: int = 0
) -> Iterable[list]:
    """
    >>> list_ = [i for i in range(100)]
    >>> for i in split_with_overlap(list_, 10,3):
    >>>     print(i)
    >>> # [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    >>> # [7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22]
    >>> # [17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32]
    >>> # [27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42]
    >>> # [37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52]
    >>> # [47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62]
    >>> # [57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72]
    >>> # [67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82]
    >>> # [77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92]
    >>> # [87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99]"""
    length = len(data)
    base_size = length // n_parts

    for i in range(n_parts):
        start = max(0, i * base_size - overlap)
        end = min(length, (i + 1) * base_size + overlap)
        yield data[start:end]



import functools
import inspect
import os
from typing import Any
import warnings


class TestModeMixin:
    """
    Mixin for switching between production and test implementations of methods
    depending on the `IS_TEST_MODE` environment variable.

    Example:
    ----------

    >>> class MyClass(TestModeMixin):
    >>>     dict_tst_mode = {"base_method": "tst_base_method"}
    >>>
    >>>     def base_method(self):
    >>>         print("PROD")
    >>>
    >>>     def tst_base_method(self):
    >>>         print("TEST")
    >>>
    >>> MyClass().base_method() # PROD
    >>> os.environ["IS_TEST_MODE"] = "1"
    >>> MyClass().base_method()  # TEST
    """

    _dict_tst_mode: dict[str, str] = {}

    @property
    def dict_tst_mode(self) -> dict[str, str]:  # pragma: no cover
        """Return mapping of base methods to their test equivalents."""

        return self._dict_tst_mode

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize mixin and patch test methods if test mode is enabled.

        Environment variables:
        - IS_TEST_MODE=1 → activates test mode
        - IS_RAISE_TEST_MODE=1 → raises exceptions instead of warnings
        """
        self._is_raise_test: int = int(os.environ.get("IS_RAISE_TEST_MODE", "0")) == 1
        super().__init__(*args, **kwargs)  # type: ignore
        self._patch_test_methods()

    def __handle_error(self, msg: str, error: type[Exception]) -> None:
        """
        Handle error when test substitution fails.

        If IS_RAISE_TEST_MODE=1 → raises an exception.
        Otherwise → emits a warning.
        """
        warnings.warn(msg, stacklevel=2)
        if self._is_raise_test:
            raise error(msg)

    def _patch_test_methods(self) -> None:
        """
        Replace production methods with test ones if IS_TEST_MODE=1.

        For every pair in dict_tst_mode:
        - Verify both methods exist and are callable.
        - Ensure both are of the same type (sync/async).
        - Dynamically replace the base method with its test equivalent.
        """
        is_test_mode = int(os.environ.get("IS_TEST_MODE", "0")) == 1
        if not is_test_mode:
            return

        for base_name, test_name in self.dict_tst_mode.items():
            base_func = getattr(self, base_name, None)
            test_func = getattr(self, test_name, None)

            if not callable(base_func):
                self.__handle_error(
                    f"[TestModeMixin]  not found base method'{base_name}' not found",
                    error=KeyError,
                )
                continue
            if not callable(test_func):
                self.__handle_error(
                    f"[TestModeMixin]  not found test method'{test_name}'",
                    error=KeyError,
                )
                continue

            if inspect.iscoroutinefunction(base_func) != inspect.iscoroutinefunction(
                test_func
            ):
                self.__handle_error(
                    f"[TestModeMixin] Not matched sync / async for methods `{test_name}` and `{base_name}`",
                    error=ValueError,
                )
                continue

            # setattr(self, base_name, functools.partial(test_func))
            # Correctly bind method to the instance (preserve `self`)
            bound_func = test_func.__get__(self, self.__class__)
            setattr(self, base_name, bound_func)



"""
The module contains a function that returns whether an event has occurred
 that has a probability of occurrence
"""

from decimal import Decimal
from math import floor
from random import uniform


def is_fate_in_awe(drop_chance: float | int | Decimal) -> bool:
    """
    Method determines whether an event has occurred that
     has a certain chance of falling out
    :param drop_chance: the probability of an event occurring
     value between[0, 1]
    :type drop_chance:  float| int| Decimal
    :return: is got this chance
    :rtype: bool
    :raises ValueError: if the value is not in the allowed range [0, 1]
    :raises TypeError: if incorrect types
    """
    value_to_compare = 1
    if not isinstance(drop_chance, (float, Decimal, int)):
        raise TypeError
    if not 0 <= drop_chance <= 1:
        raise ValueError(
            "Invalid variable value `drop_chance` value must be in between [0, 1]"
        )
    if drop_chance > 0.5:
        # for correct calculation
        value_to_compare = 0
    return floor(uniform(0, 1 / float(drop_chance))) == value_to_compare



"""
Setup orch_serv
"""
import os
import pathlib
import re
from typing import Optional

from setuptools import setup

LIB_NAME = "my_utilities"
HERE = pathlib.Path(__file__)
DESCRIPTION = """
Library with helper functions and classes
"""


def get_version() -> Optional[str]:
    """
      Method for getting the version of the library from the init file
    :requirements: version must be specified separately
        :good: __version__ = '0.0.1'
        :bad: __version__, __any_variable__ = '0.0.1', 'any_value'
    :return: version lib
    """
    root_lib = pathlib.Path(__file__).parent / LIB_NAME
    txt = (root_lib / "__init__.py").read_text("utf-8")
    txt = txt.replace("'", '"')
    try:
        version = re.findall(r'^__version__ = "([^"]+)"\r?$', txt, re.M)[0]
        return version
    except IndexError:
        raise RuntimeError("Unable to determine version.") from Exception


def get_packages():
    """
    Method for getting packages used in the lib
    """
    ignore = ["__pycache__"]

    list_sub_folders_with_paths = [
        x[0].replace(os.sep, ".")
        for x in os.walk(LIB_NAME)
        if x[0].split(os.sep)[-1] not in ignore
    ]
    return list_sub_folders_with_paths


def get_long_description() -> str:
    """
    Get long description for setup function
    :return:  description
    """
    path_readme_pypi = pathlib.Path("README_PYPI.md")
    path_readme = pathlib.Path("README_PYPI.md")
    if path_readme_pypi.exists():
        with open(path_readme_pypi, "r", encoding="utf8") as file:
            return file.read()
    elif path_readme.exists():
        with open(path_readme, "r", encoding="utf8") as file:
            return file.read()
    else:
        return DESCRIPTION


setup(
    name=LIB_NAME,
    version=get_version(),
    description=DESCRIPTION,
    author="Denis Shchutkiy",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author_email="denisshchutskyi@gmail.com",
    url="https://github.com/Shchusia/MyUtils",
    packages=get_packages(),
    keywords=["pip", LIB_NAME],
    python_requires=">=3.7",
    install_requires=["pydantic==1.9.0"],
)
